﻿using System;
using System.IO;
using System.Security.Cryptography;
using System.Text;
using System.Web;
using Org.BouncyCastle.Crypto.Parameters;
using Org.BouncyCastle.Security;

namespace TestDameKnihu {
	public enum VoucherStatuses : byte {
		NotPaidYet = 0, Active = 1, Used = 2, Deactivated = 3, NotYetActive = 4, Expired = 5
	};
	public class DameKnihu {
		private const string ApiUrl = "https://www.dameknihu.cz/API-v1";

		#region Certificates
		private static RSAParameters _privateCertificate;
		private static RSAParameters _publicCertificate;
		public static bool LoadCertificates(string privateCertificateFile, string publicCertificateFile) {
			try {
				_privateCertificate = LoadPrivateKey(privateCertificateFile);
				_publicCertificate = LoadPublicKey(publicCertificateFile);
				return true;
			} catch (Exception ex) {
				//Logger.LogMsgT("Failed to load DameKnihu certificates: " + ex.ToString());
				return false;
			}
		}

		private static RSAParameters LoadPrivateKey(string path) {
			using (var sr = File.OpenText(path)) {
				var pemReader = new Org.BouncyCastle.Utilities.IO.Pem.PemReader(sr);
				var pem = pemReader.ReadPemObject();
				var privateKey = PrivateKeyFactory.CreateKey(pem.Content);

				return DotNetUtilities.ToRSAParameters(privateKey as RsaPrivateCrtKeyParameters);
			}
		}

		private static RSAParameters LoadPublicKey(string path) {
			using (var sr = File.OpenText(path)) {
				var pemReader = new Org.BouncyCastle.Utilities.IO.Pem.PemReader(sr);
				var pem = pemReader.ReadPemObject();
				var publicKey = PublicKeyFactory.CreateKey(pem.Content);

				return DotNetUtilities.ToRSAParameters(publicKey as RsaKeyParameters);
			}
		}

		private const int ssl_padding_length = 42;
		private const int ssl_signature_length = 512;

		private static byte[] Encode(string data, byte[] signature) {
			using (var provider = new RSACryptoServiceProvider()) {
				provider.ImportParameters(_publicCertificate);

				var msgData = Encoding.UTF8.GetBytes(data);
				byte[] dataToEncode = new byte[msgData.Length + signature.Length];
				msgData.CopyTo(dataToEncode, 0);
				signature.CopyTo(dataToEncode, msgData.Length);

				var dataLength = dataToEncode.Length;
				var maxLength = (int)Math.Floor(((decimal)provider.KeySize) / 8 - ssl_padding_length);


				if (dataLength > maxLength) {
					var stream = new MemoryStream();
					var iterations = (int)Math.Ceiling((decimal)dataToEncode.Length / maxLength);
					for (var i = 0; i < iterations; i++) {
						byte[] block;
						if (i < iterations - 1) {
							block = new byte[maxLength];
						} else {
							block = new byte[dataLength - maxLength * i];
						}

						// copy the required number of bytes from the array of bytes to encrypt to our worker array
						Buffer.BlockCopy(dataToEncode, maxLength * i, block, 0, block.Length);

						// encrypt the current worker array block of bytes
						var encryptedBytes = provider.Encrypt(block, true);
						stream.Write(encryptedBytes, 0, encryptedBytes.Length);
					}
					return stream.ToArray();
				} else {
					return provider.Encrypt(dataToEncode, true);
				}
			}
		}

		private static byte[] Decrypt(byte[] data) {
			using (var provider = new RSACryptoServiceProvider()) {
				provider.ImportParameters(_privateCertificate);


				var dataLength = data.Length;
				var maxLength = (int)Math.Floor(((decimal)provider.KeySize) / 8);

				if (dataLength > maxLength) {
					var stream = new MemoryStream();
					var iterations = (int)Math.Ceiling((decimal)data.Length / maxLength);
					for (var i = 0; i < iterations; i++) {
						byte[] block;
						if (i < iterations - 1) {
							block = new byte[maxLength];
						} else {
							block = new byte[dataLength - maxLength * i];
						}

						// copy the required number of bytes from the array of bytes to encrypt to our worker array
						Buffer.BlockCopy(data, maxLength * i, block, 0, block.Length);

						// encrypt the current worker array block of bytes
						var decryptedBytes = provider.Decrypt(block, true);
						stream.Write(decryptedBytes, 0, decryptedBytes.Length);
					}
					return stream.ToArray();
				} else {
					return provider.Decrypt(data, true);
				}
			}
		}

		private static byte[] Sign(string data) {
			using (var provider = new RSACryptoServiceProvider()) {
				provider.ImportParameters(_privateCertificate);

				var signature = provider.SignData(UTF8Encoding.UTF8.GetBytes(data), SHA512.Create());
				return signature;
			}
		}

		private static bool Verify(string data, byte[] signature) {
			using (var provider = new RSACryptoServiceProvider()) {
				provider.ImportParameters(_publicCertificate);

				return provider.VerifyData(UTF8Encoding.UTF8.GetBytes(data), SHA512.Create(), signature);
			}
		}

		private static string Decode(byte[] data) {
			var decryptedData = Decrypt(data);
			byte[] signature = new byte[ssl_signature_length];
			Buffer.BlockCopy(decryptedData, decryptedData.Length - ssl_signature_length, signature, 0, ssl_signature_length);

			var json = Encoding.UTF8.GetString(decryptedData, 0, decryptedData.Length - ssl_signature_length);

			byte[] goodSignature = Sign(json);

			if (Verify(json, signature)) {
				return json;
			} else {
				return null;
			}
		}
		#endregion

		private static System.Text.RegularExpressions.Regex reValidCode = new System.Text.RegularExpressions.Regex("^DK-?[a-zA-Z0-9]{4}-?[a-zA-Z0-9]{4}$");
		public static bool ValidateFormat(string code) {
			return reValidCode.IsMatch(code);
		}

		public static int Pobocka { get; set; }

		public static VoucherInfo Validate(string code, string email) {
			if (!ValidateFormat(code)) {
				return new VoucherInfo { Accepted = false, Message = "Nesprávný formát kódu poukazu.", Status = VoucherStatuses.Deactivated };
			}

			var request = new OveritRequest(code, email);
			var requestJson = Utils.SerializeToJSON(request);

			var signature = Sign(requestJson);
			var encoded = Encode(requestJson, signature);

			var (httpCode, encrypted) = Utils.PostFormToURLRaw(ApiUrl, "application/x-www-form-urlencoded", "data=" + HttpUtility.UrlEncode(encoded));
			if (httpCode == System.Net.HttpStatusCode.OK) {
				// dostali jsme JSON

				var resultJson = Decode(encrypted);
				if (resultJson == null) {
					return new VoucherInfo { Accepted = false, Message = "Nesprávný podpis dotazu.", Status = VoucherStatuses.Deactivated };
				}

				var result = Newtonsoft.Json.JsonConvert.DeserializeObject<Response>(resultJson);

				switch (result.status) {
					case States.TryLater:
						return new VoucherInfo { Accepted = false, Message = "V tuto chvíli není možné poukazy z DámeKnihu validovat. Zkuste to prosím později.", Status = VoucherStatuses.Deactivated };
					case States.NotFound:
						return new VoucherInfo { Accepted = false, Message = "Nesprávný kód poukazu.", Status = VoucherStatuses.Deactivated };
					case States.InvalidCode:
						return new VoucherInfo { Accepted = false, Message = "Nesprávný formát kódu poukazu.", Status = VoucherStatuses.Deactivated };
					case States.AlreadyUsed:
						return new VoucherInfo { Accepted = false, Message = "Tento poukaz už byl využit.", Status = VoucherStatuses.Used };
					case States.Expired:
						return new VoucherInfo { Accepted = false, Message = "Platnost poukazu už vypršela.", Status = VoucherStatuses.Expired, ValidTo = result.data.datum_platnosti_val };
					case States.BlockedByOther:
						return new VoucherInfo { Accepted = false, Message = "Poukaz je v tuto chvíli blokován na jiné prodejně.", Status = VoucherStatuses.Deactivated };
					case States.ReservedForMe:
					case States.Available: // mělo to zarezervovat, ale budeme doufat
						return new VoucherInfo {
							Accepted = true,
							Message = result.data.datum_platnosti_txt + " Hodnota: " + result.data.hodnota_txt,
							Value = result.data.hodnota,
							ValidTo = result.data.datum_platnosti_val,
							Status = VoucherStatuses.Active
						};
					case States.Used:
						//LogException("DameKnihu Overit", new Exception($"Při ověřování to poukaz použilo. WTF?!? Code: {code}"));
						return new VoucherInfo { Accepted = false, Message = "Tento poukaz už byl využit.", Status = VoucherStatuses.Used };
					default:
						return new VoucherInfo { Accepted = false, Message = "Došlo k chybě při pokusu o validaci poukazu. Zkuste to prosím později", Status = VoucherStatuses.Deactivated };
				}

			} else {
				//LogException("DameKnihu Overit", new Exception($"Code: {code}, HTTP: {httpCode}, Result: {encrypted}"));
				return new VoucherInfo { Accepted = false, Message = $"Došlo k chybě při ověřování poukazu. ({encrypted})", Status = VoucherStatuses.Deactivated };
			}
		}

		public static VoucherInfo Use(string code, string email, string varnum) {
			if (!ValidateFormat(code)) {
				return new VoucherInfo { Accepted = false, Message = "Nesprávný formát kódu poukazu.", Status = VoucherStatuses.Deactivated };
			}

			var request = new CerpatRequest(code, email, varnum);
			var requestJson = Utils.SerializeToJSON(request);

			var signature = Sign(requestJson);
			var encoded = Encode(requestJson, signature);

			var (httpCode, encrypted) = Utils.PostFormToURLRaw(ApiUrl, "application/x-www-form-urlencoded", "data=" + HttpUtility.UrlEncode(encoded));
			if (httpCode == System.Net.HttpStatusCode.OK) {
				// dostali jsme JSON
				var resultJson = Decode(encrypted);
				if (resultJson == null) {
					return new VoucherInfo { Accepted = false, Message = "Nesprávný podpis dotazu.", Status = VoucherStatuses.Deactivated };
				}

				var result = Newtonsoft.Json.JsonConvert.DeserializeObject<Response>(resultJson);

				switch (result.status) {
					case States.TryLater:
						return new VoucherInfo { Accepted = false, Message = "V tuto chvíli není možné poukazy z DámeKnihu validovat. Zkuste to prosím později.", Status = VoucherStatuses.Deactivated };
					case States.NotFound:
						return new VoucherInfo { Accepted = false, Message = "Nesprávný kód poukazu.", Status = VoucherStatuses.Deactivated };
					case States.InvalidCode:
						return new VoucherInfo { Accepted = false, Message = "Nesprávný formát kódu poukazu.", Status = VoucherStatuses.Deactivated };
					case States.AlreadyUsed:
						return new VoucherInfo { Accepted = false, Message = "Tento poukaz už byl využit.", Status = VoucherStatuses.Used };
					case States.Expired:
						return new VoucherInfo { Accepted = false, Message = "Platnost poukazu už vypršela.", Status = VoucherStatuses.Expired, ValidTo = result.data.datum_platnosti_val };
					case States.BlockedByOther:
						return new VoucherInfo { Accepted = false, Message = "Poukaz je v tuto chvíli blokován na jiné prodejně.", Status = VoucherStatuses.Deactivated };
					case States.ReservedForMe:
					case States.Available: // mělo to zarezervovat, ale budeme doufat
										   //LogException("DameKnihu Cerpat", new Exception($"Při ověřování to poukaz jen zarezervovalo. WTF?!? Code: {code}"));
						return new VoucherInfo { Accepted = false, Message = "Poukaz se nepodařilo označit jako použitý.", Status = VoucherStatuses.Deactivated };
					case States.Used:
						return new VoucherInfo {
							Accepted = true,
							Message = result.data.datum_platnosti_txt + " Hodnota: " + result.data.hodnota_txt,
							Value = result.data.hodnota,
							ValidTo = result.data.datum_platnosti_val,
							Status = VoucherStatuses.Active
						};
					default:
						return new VoucherInfo { Accepted = false, Message = "Došlo k chybě při pokusu o validaci poukazu. Zkuste to prosím později", Status = VoucherStatuses.Deactivated };
				}

			} else {
				//LogException("DameKnihu Cerpat", new Exception($"Code: {code}, HTTP: {httpCode}, Result: {encrypted}"));
				return new VoucherInfo { Accepted = false, Message = $"Došlo k chybě při čerpání poukazu. ({encrypted})", Status = VoucherStatuses.Deactivated };
			}
		}

		public class VoucherInfo {
			public bool Accepted { get; set; }
			public VoucherStatuses Status { get; set; }
			public DateTime? ValidTo { get; set; }
			public decimal Value { get; set; }
			public string Message { get; set; }
		}

		private class OveritRequest {
			public string akce { get; set; }
			public int pobocka { get; set; }
			public string kod { get; set; }
			public string uzivatel { get; set; }

			public OveritRequest(string code, string email) {
				akce = "overit";
				pobocka = DameKnihu.Pobocka;
				kod = code;
				uzivatel = email;
			}
		}

		private class CerpatRequest : OveritRequest {
			public string poznamka { get; set; }

			public CerpatRequest(string code, string email, string orderId) : base(code, email) {
				akce = "cerpat";
				poznamka = "Varnum: " + orderId;
			}
		}

		private enum States : byte {
			TryLater = 1, NotFound, InvalidCode, AlreadyUsed, Expired, BlockedByOther, ReservedForMe, Used, Available, InvalidStatus
		}

		private class Response {
			public char stav { get; set; } // jedno písmeno
			/*
			F	Chyba - ověření kódu nepovoleno, zkuste později!
			N	Chyba - poukaz nenalezen
			E	Chyba - formát kódu neodpovídá!
			U	Poukaz byl vyčerpán dříve, nelze čerpat
			X	Platnost poukazu vypršela, nelze čerpat
			B	Poukaz má momentálne rezervovaný jiná pobočka
			R	Poukaz je platný, aktuálně rezervovaný pro vás
			P	Poukaz byl úspěšně vyčerpán (pouze pro akce='cerpat')
			A	Poukaz je platný (pouze pro front-end portálu)
			*/

			private States _status;
			public States status
			{
				get {
					if (_status == 0) {
						switch (stav) {
							case 'F':
								_status = States.TryLater;
								break;
							case 'N':
								_status = States.NotFound;
								break;
							case 'E':
								_status = States.InvalidCode;
								break;
							case 'U':
								_status = States.AlreadyUsed;
								break;
							case 'X':
								_status = States.Expired;
								break;
							case 'B':
								_status = States.BlockedByOther;
								break;
							case 'R':
								_status = States.ReservedForMe;
								break;
							case 'P':
								_status = States.Used;
								break;
							case 'A':
								_status = States.Available;
								break;
							default:
								_status = States.InvalidStatus;
								break;
						}
					}
					return _status;
				}
			}
			public string text { get; set; }

			public Data data { get; set; }
		}

		private class Data {
			public decimal hodnota { get; set; }
			public string hodnota_txt { get; set; }
			public string stav_txt { get; set; }

			public long? datum_blokace { get; set; }
			public DateTime? datum_blokace_val { get { return datum_blokace.HasValue ? DateTimeOffset.FromUnixTimeSeconds(datum_blokace.Value).DateTime : null; } }
			public string datum_blokace_txt { get; set; }

			public long? datum_platnosti { get; set; }
			public DateTime? datum_platnosti_val { get { return datum_platnosti.HasValue ? DateTimeOffset.FromUnixTimeSeconds(datum_platnosti.Value).DateTime : null; } }
			public string datum_platnosti_txt { get; set; }

			public long? datum_cerpani { get; set; }
			public DateTime? datum_cerpani_val { get { return datum_cerpani.HasValue ? DateTimeOffset.FromUnixTimeSeconds(datum_cerpani.Value).DateTime : null; } }
			public string datum_cerpani_txt { get; set; }

			public int prodejce_cerpani { get; set; }
			public int pobocka_cerpani { get; set; }
		}
	}
}
